#include "ColorDetector.h"

ColorDetector::ColorDetector ()
{
  _blobs = new cvb::CvBlobs();
}

bool ColorDetector::init (const int width, const int height, const int imgs, unsigned int& add, const char* fname, const char* chapter)
{
  _offset = imgs;
  add = 2;

  for (int i=0; i<3; i++)
    _lasts.push_back(TapirObject(0,0,0,0));

  ValueParser vp(fname, CHAPTER "Detector");

  // min and max blob size
  _area[0] = 80;
  _area[1] = 100000;
  vp.get("area", _area, 2);
  if (_area[0] > _area[1]) {
    int tmpi = _area[0];
    _area[0] = _area[1];
    _area[1] = tmpi;
  }

  vp.get("color", _color, 3);
  vp.get("wide",  _wide,  3);

  vp.get("saturation", _saturation, 0);
  vp.get("brightness", _brightness, 0);

  _img       = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, 3);
  _label     = cvCreateImage(cvSize(width,height), IPL_DEPTH_LABEL, 1);
  _hsvframe  = cvCreateImage(cvSize(width,height), 8, 3); // Image in HSV color space
  _threshy   = cvCreateImage(cvSize(width,height), 8, 1);
  _convert   = cvCreateImage(cvSize(width,height), 8, 3);
  _converted = cvCreateImage(cvSize(width,height), 8, 3);

  REGISTER_KEY(ColorDetector, '4', "increase RGB R");
  REGISTER_KEY(ColorDetector, 'r', "decrease RGB R");
  REGISTER_KEY(ColorDetector, '5', "increase RGB G");
  REGISTER_KEY(ColorDetector, 't', "decrease RGB G");
  REGISTER_KEY(ColorDetector, '6', "increase RGB B");
  REGISTER_KEY(ColorDetector, 'y', "decrease RGB B");
  REGISTER_KEY(ColorDetector, '7', "increase RGB R-range");
  REGISTER_KEY(ColorDetector, 'u', "decrease RGB R-range");
  REGISTER_KEY(ColorDetector, '8', "increase RGB G-range");
  REGISTER_KEY(ColorDetector, 'i', "decrease RGB G-range");
  REGISTER_KEY(ColorDetector, '9', "increase RGB B-range");
  REGISTER_KEY(ColorDetector, 'o', "decrease RGB B-range");

  ///< \todo this does not belong in a detector; neither really in an imagesource, so we might add a "preprocessor" module type
  REGISTER_KEY(ColorDetector, 'b', "select parameter brightness");
  REGISTER_KEY(ColorDetector, 's', "select parameter saturation");
  REGISTER_KEY(ColorDetector, 'e', "de-select parameter");
  REGISTER_KEY(ColorDetector, 'p', "print selected parameter");
  REGISTER_EXTENDED_KEY(ColorDetector, "ARROW_UP",    27,91,65, "increase selected parameter");
  REGISTER_EXTENDED_KEY(ColorDetector, "ARROW_DOWN",  27,91,66, "decrease selected parameter");
  REGISTER_EXTENDED_KEY(ColorDetector, "ARROW_RIGHT", 27,91,67, "increase selected parameter");
  REGISTER_EXTENDED_KEY(ColorDetector, "ARROW_LEFT",  27,91,68, "decrease selected parameter");

  return true;
}

void ColorDetector::deinit ()
{
  cvReleaseImage(&_img);
  cvReleaseImage(&_label);
  cvReleaseImage(&_hsvframe);
  cvReleaseImage(&_threshy);
  cvReleaseImage(&_convert);
  cvReleaseImage(&_converted);
}

void ColorDetector::notify (const long int key, const char* params)
{
  switch (key) {
  case '4':
    _color[0] += 1;
    break;
  case '5':
    _color[1] += 1;
    break;
  case '6':
    _color[2] += 1;
    break;
  case 'r':
    _color[0] -= 1;
    break;
  case 't':
    _color[1] -= 1;
    break;
  case 'z':
    _color[2] -= 1;
    break;
  case '7':
    _wide[0] += 1;
    break;
  case '8':
    _wide[1] += 1;
    break;
  case '9':
    _wide[2] += 1;
    break;
  case 'u':
    _wide[0] -= 1;
    break;
  case 'i':
    _wide[1] -= 1;
    break;
  case 'o':
    _wide[2] -= 1;
    break;

  case 'b' :
    IOUT(" Change brightness mode");
    _state = BRIGHTNESS;
    break;

  case 's' :
    IOUT(" Change saturation mode");
    _state = SATURATION;
    break;

  case 'e' :
    IOUT(" OFF");
    _state = OFF;
    break;

  case 'p':
    switch (_state) {
    case BRIGHTNESS :
      IOUT(" brightness = " << _brightness);
      break;

    case SATURATION :
      IOUT(" saturation = " << _saturation);
      break;

    case OFF:
    default :
      cout<<endl<<"Color : "<<endl;
      cout<<      "        ("<<_color[0]<<", "<<_color[1];
      cout<<", "<<_color[2]<<")"<<endl;
      cout<<endl<<"Wide  : "<<endl;
      cout<<      "        ("<<_wide[0]<<", "<<_wide[1];
      cout<<", "<<_wide[2]<<")"<<endl;

      break;
    }
    break;

  case 4283163 : // UP key
    switch (_state) {
    case BRIGHTNESS :
      _brightness += 5;
      break;

    case SATURATION :
      _saturation += 2;
      break;

    case OFF:
    default:
      break;
    }
    break;
  case 4348699 : // DOWN key
    switch (_state) {
    case BRIGHTNESS :
      _brightness -= 5;
      break;

    case SATURATION :
      _saturation -= 2;
      break;

    case OFF:
    default:
      break;

    }
    break;

  case 4414235 : // RIGHT key
    switch (_state) {
    case BRIGHTNESS :
      _brightness += 5;
      break;

    case SATURATION :
      _saturation += 2;
      break;

    case OFF:
    default:
      break;
    }
    break;
  case 4479771 : // LEFT key
    switch (_state) {
    case BRIGHTNESS :
      _brightness -= 5;
      break;

    case SATURATION :
      _saturation -= 2;
      break;

    case OFF:
    default:
      break;
    }

  default:
    return;
  }
}

bool ColorDetector::detect (IplImage** imgs, std::vector<TapirObject>& objects)
{
  if (imgs[0] == 0) return true;

  // Make a copy of the image and adjust color and brightness
  cvCopy(imgs[0], _img);
  enhance_BGR_colors(_img);

  // Changing the color space
  cvCvtColor(_img, _hsvframe, CV_BGR2HSV);

  // Thresholding the frame for the given range of colors
  cvInRangeS(_hsvframe, get_min_color(), get_max_color(), _threshy);

  // Filtering the frame
  cvSmooth(_threshy, _threshy, CV_MEDIAN, 7, 7);

  // Finding the blobs
  cvLabel(_threshy, _label, *_blobs);

  // Filtering the blobs
  cvFilterByArea(*_blobs, _area[0], _area[1]);

  objects.clear();
  cvb::CvBlobs::iterator i;
  cvb::CvBlob *b;
  for (i=_blobs->begin(); i!=_blobs->end(); i++) {
    b = (*i).second;
    objects.push_back(TapirObject(b->centroid.x, b->centroid.y, b->area));
  }

  // generate output image
  cvRenderBlobs(_label, *_blobs, _img, _img, CV_BLOB_RENDER_CENTROID | CV_BLOB_RENDER_BOUNDING_BOX);

  imgs[_offset]   = _img;
  imgs[_offset+1] = _threshy;

  return true;
}

CvScalar ColorDetector::get_min_color ()
{
  return cvScalar(saturate(_color[0]-_wide[0]),
                  saturate(_color[1]-_wide[1]),
                  saturate(_color[2]-_wide[2]));
}

CvScalar ColorDetector::get_max_color ()
{
  return cvScalar(saturate(_color[0]+_wide[0]),
                  saturate(_color[1]+_wide[1]),
                  saturate(_color[2]+_wide[2]));
}

int ColorDetector::saturate (int x)
{
  return x > 255 ? 255 : x < 0 ? 0 : x;
}

bool ColorDetector::enhance_BGR_colors (IplImage* img)
{
  if (_saturation!=0 || _brightness!=0) {
    cvCvtColor(img, _convert, CV_BGR2HSV);

    int height    = img->height;
    int width     = img->width;
    int step      = img->widthStep;
    int channels  = img->nChannels;
    uchar* data = (uchar *)_convert->imageData;

    for (int h=0; h<height; h++) {
      for (int j=0; j<width; j++) {
        data[h*step+j*channels+1] = saturate(data[h*step+j*channels+1]+_saturation);
        data[h*step+j*channels+2] = saturate(data[h*step+j*channels+2]+_brightness);
      }
    }

    cvCvtColor(_convert, _converted, CV_HSV2BGR);

    uchar* imgdata = (uchar *)img->imageData;
    uchar* convdata = (uchar *)_converted->imageData;
    for (int h=0; h<height; h++) {
      for (int j=0; j<width; j++) {
        for (int k=0; k<channels; k++)
          imgdata[h*step+j*channels+k] = saturate(convdata[h*step+j*channels+k]);
      }
    }

    return true;
  }

  return false;
}

REGISTER(Detector, ColorDetector, "Detects objects of a given color");
