#ifndef _THRESHOLD_DETECTOR_
#define _THRESHOLD_DETECTOR_

#include "detector.h"
#include "cvblob.h"

/** This detector gives a discretized description of the image.
  * In each block of the discretized image, it calculates the mean color and
  * returns the closest given color to each blocks.
  * An object is returned for each block with .x,.y,and .size as color index. The index is the position
  * of the color in the given color array (0|1|2|...)

  * A detetion frame can be adjusted on the wanted region of the image with l,u, and arrow keys.
  * In order to do some testing, the get_block_color mode enable to get distance values between
  * a given mean color of a block and given colors for detection.
  *
  * Params :
  * \li  grid_x - int, Number of blocks in x
  * \li  grid_y - int, Number of blocks in y (default = grid_x)
  * \li  x_min  - int, Lower limit of the detection frame on the image
  * \li  x_max  - int, Upper limit of the detection frame on the image
  * \li  y_min  - int, Left limit of the detection frame on the image
  * \li  y_max  - int, Right limit of the detection frame on the image
  * \li  print_block_color - bool, Print the mean color of the block on the raw image
  * \li  get_block_color   - bool, Enable the get_block_color on start
  * \li  get_color_x       - int,  Initial x position of get_block_color block
  * \li  get_color_y       - int,  Initial y position of get_block_color block
  * \li  saturation        - int,  Modify saturation of the image
  * \li  brightness        - int,  Modify brightness of the image
  * \li  colors            - int,int,int|int,int,int|..., given colors to detect in RGB form
  *
  * Keys:
  * \li  l : change minimum (left,up) bounds mode
  *         left  : y_min -= 2
  *         right : y_min += 2
  *         up    : y_min -= 2
  *         down  : y_min += 2
  *         p     : print x_min,x_max,y_min,y_max
  * \li  u : change maximum (right,down) bounds mode
  *         left  : y_max -= 2
  *         right : y_max += 2
  *         up    : x_max -= 2
  *         down  : x_max += 2
  *         p     : print x_min,x_max,y_min,y_max
  * \li  b : change brightness mode
  *         left, down : -5 to brightness
  *         right, up  : +5 to brightness
  *         p          : print brightness value
  * \li  s : change saturation mode
  *         left, down : -2 to saturation
  *         right, up  : +2 to saturation
  *         p          : print saturation value
  * \li  c : get_block_color mode
  *         left  : move get_block_color block to the left
  *         right : move get_block_color block to the right
  *         up    : move get_block_color block up
  *         down  : move get_block_color block down
  *         p     : print get_block_color block position and comparisons
  *                 with mean color of the block and the given color
  *
  * @ingroup DETECTOR
  * @author Xavier Bouthillier */
class ColoredGridDetector : public Tapir::Detector
{
  enum state_t {OFF,LOWER,UPPER,COLOR,BRIGHTNESS,SATURATION};

public:
  ColoredGridDetector ();
  ~ColoredGridDetector ();
  bool detect (IplImage** imgs, std::vector<TapirObject>& objects);
  bool detect_test (IplImage* raw, IplImage*& detect, std::vector<TapirObject>& objects);
  bool init (const int width, const int height, const int imgs, unsigned int& add, const char* fname, const char* chapter);
  void deinit ();
  void notify (const long int cmd, const char* params);

protected:

  IplImage *_detect, *_raw, *_convert, *_converted;
  int _offset;
  int  _saturation, _brightness;
  int _grid_x,_grid_y,_block_height,_block_width;
  int _x_min, _x_max;
  int _y_min, _y_max;
  bool _get_block_color,_print_block_color;
  int _get_color_x,_get_color_y;

  state_t _state;
  int _channels;
  int *_blocks;
  std::vector<CvScalar> _colors;

  int distance_to_color(int R1, int G1, int B1, int R2, int G2, int B2);

  /** Draws a grid of the given color on the given image.
    * \param img image to draw on
    * \param color color to draw grid in */
  bool draw_grid(IplImage* img, CvScalar color);

  /** Enhances colors via saturation and brightness.
    * \param img image to enhance */
  bool enhance_BGR_colors(IplImage* img);

  /** Prints distance results between the wanted colors
    * and the actual color of the block. */
  bool get_block_color();

  /** Draws the detected color on a block. */
  bool draw_get_block_color();

  /** Sums color of each block region in the histogram _blocks.
   *  [x][y][color] */
  bool get_grid(IplImage* img);

  int get_block_index(int x, int y, int z);
  int get_block_value(int x, int y, int z);

  /** Reads color from a string and returns it as a CvScalar
    * \param colors string containing colors in format "R,G,B"
    * \return color object */
  CvScalar read_color_str(std::string colors);

private:

  int saturate(int value);
};

#endif
