#ifndef _MAZE_DETECTOR_
#define _MAZE_DETECTOR_

#include "ColoredGridDetector.h"
#include "cvblob.h"

using namespace Tapir;

/** This detector gives a discretezied description of a maze.
  * The detector calculate the mean color of each blocks and assign them to ground or walls. 
  * Colors can be added with the colors parameter as 0,0,0|50,50,50 for example to assign objects
  * such a goals, starting points or traps. 
  * An object is returned for each block with .x,.y,and .size as color index. The index is the position
  * of the color in the given color array (0|1|2|...) + 2, because 0 is the index of the ground color 
  * and 1 is the index of walls color. 
  * 
  * A detetion frame can be adjusted on the wanted region of the image with l,u, and arrow keys.
  * In order to do some testing, the get_block_color mode enable to get distance values between
  * a given mean color of a block and given colors for detection. 
  *
  * Params :
  * \li  grid_x - int, Number of blocks in x 
  * \li  grid_y - int, Number of blocks in y (default = grid_x)
  * \li  x_min  - int, Lower limit of the detection frame on the image
  * \li  x_max  - int, Upper limit of the detection frame on the image
  * \li  y_min  - int, Left limit of the detection frame on the image
  * \li  y_max  - int, Right limit of the detection frame on the image
  * \li  print_block_color - bool, Print the mean color of the block on the raw image 
  * \li  get_block_color   - bool, Enable the get_block_color on start
  * \li  get_color_x       - int,  Initial x position of get_block_color block
  * \li  get_color_y       - int,  Initial y position of get_block_color block
  * \li  saturation        - int,  Modify saturation of the image
  * \li  brightness        - int,  Modify brightness of the image
  * \li  colors            - int,int,int|int,int,int|..., given colors to detect in RGB form
  * \li  ground            - int,int,int, color of the ground (default = 155,155,155)(white)
  * \li  walls             - int,int,int, color of the walls (default = 0,0,0)(black) 
  * \li  max_distance_to_color - int, maximal distance to color before it is counted as ground of walls
  *
  * Keys:
  * \li  l : change minimum (left,up) bounds mode 
  *         left  : y_min -= 2
  *         right : y_min += 2
  *         up    : y_min -= 2
  *         down  : y_min += 2
  *         p     : print x_min,x_max,y_min,y_max
  * \li  u : change maximum (right,down) bounds mode 
  *         left  : y_max -= 2
  *         right : y_max += 2
  *         up    : x_max -= 2
  *         down  : x_max += 2
  *         p     : print x_min,x_max,y_min,y_max
  * \li  b : change brightness mode
  *         left, down : -5 to brightness
  *         right, up  : +5 to brightness
  *         p          : print brightness value
  * \li  s : change saturation mode 
  *         left, down : -2 to saturation
  *         right, up  : +2 to saturation
  *         p          : print saturation value
  * \li    c : get_block_color mode
  *         left  : move get_block_color block to the left
  *         right : move get_block_color block to the right
  *         up    : move get_block_color block up
  *         down  : move get_block_color block down
  *         p     : print get_block_color block position and comparisons 
  *                 with mean color of the block and the given color
  *
  * @ingroup DETECTOR
  * @author Xavier Bouthillier */
class MazeDetector : public ColoredGridDetector 
{
 public:
  MazeDetector ();
  bool init (const int width, const int height, const int imgs, unsigned int& add, const char* fname, const char* chapter);
  void deinit ();
  bool detect (IplImage** imgs, std::vector<TapirObject>& objects);

 private:
  CvScalar _ground_color, _wall_color;
  int _max_dist_color;
};

#endif
