#include "HistogramDetector.h"

#define SETSTRING(target,source) { target = new char[strlen(source)+1]; strncpy(target, source, strlen(source)); target[strlen(source)]='\0'; }

HistogramDetector::HistogramDetector ()
{
  _blobs = new cvb::CvBlobs();
  _neg = new HistogramFeature::Histogram();
  _pos = new HistogramFeature::Histogram();
  _dif = new HistogramFeature::Histogram();
  _first = true;
  _lock = false;
}

bool HistogramDetector::init (const int width, const int height, const int imgs, unsigned int& add, const char *fname, const char* chapter)
{
  ValueParser vp(fname, CHAPTER "Detector");

  SETSTRING(_filename, "histograms.xml");
  char filename[255];
  if (vp.get("filename", filename, 255) > 0) {
    _filename = new char[255];
    strcpy(_filename,filename);
  }

  // try to load from file?
  vp.get("autoload", _autoload, false);

  // initial examples are needed only when not loading
  if (_autoload) _ex[0] = 1;
  else {
    vp.get("examples_init", _ex[0], 50);
    if (_ex[0] < 1) {
      WOUT("Setting examples_init to minimum of 1.");
      _ex[0] = 1;
    }
  }

  // examples when pressing +
  vp.get("examples_renew", _ex[1], 1);
  if (_ex[1] < 1) {
    WOUT("Setting examples_renew to minimum of 1.");
    _ex[1] = 1;
  }

  // min and max blob size
  _area[0] = 80;
  _area[1] = 100000;
  vp.get("area", _area, 2);
  if (_area[0] > _area[1]) {
    int tmpi = _area[0];
    _area[0] = _area[1];
    _area[1] = tmpi;
  }

  _detect = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 1);
  _label  = cvCreateImage(cvSize(width, height), IPL_DEPTH_LABEL, 1);
  _canvas = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);
  _updates = _ex[0];

  // remember index to write our two own images into
  _imgoffset = imgs;
  add = 2;

  REGISTER_KEY(HistogramDetector, ' ', "Start or stop adding frames to the background");
  REGISTER_KEY(HistogramDetector, '+', "Add a single frame to the background");

  std::stringstream ss;
  ss << "Load histogram saved as \"" << _filename << "\"";
  REGISTER_KEY(HistogramDetector, 'l', ss.str().c_str());
  ss.str("");
  ss << "Save histogram as \"" << _filename << "\"";
  REGISTER_KEY(HistogramDetector, 's', ss.str().c_str());

  return true;
}

void HistogramDetector::deinit ()
{
  cvReleaseImage(&_detect);
  cvReleaseImage(&_label);
  cvReleaseImage(&_canvas);
}

bool HistogramDetector::load ()
{
  CvFileStorage *storage = cvOpenFileStorage(_filename, 0, CV_STORAGE_READ);
  if (storage == 0) {
    EOUT("Failed to load histograms, file storage \"" << _filename << "\" does not exist.");
    return false;
  } 
  CvFileNode *node;
#define LOADCVOBJECT(name,target) \
  node = cvGetFileNodeByName(storage,0,name); \
  target = (CvHistogram*)cvRead(storage, node, new CvAttrList());
  LOADCVOBJECT("neg",_neg->hist);
  LOADCVOBJECT("pos",_pos->hist);
  LOADCVOBJECT("dif",_dif->hist);
  cvReleaseFileStorage(&storage);
  IOUT("Loaded histograms from location \"" << _filename << "\".");
  return true;
}

void HistogramDetector::notify (const long int cmd, const char* params)
{
  switch (cmd) {
    case ' ':
      _lock = !_lock;
      break;
    case '+':
      _updates = _ex[1];
      break;
    case 's':
    {
      CvFileStorage *storage = cvOpenFileStorage(string(_filename).c_str(), 0, CV_STORAGE_WRITE_TEXT);
      cvWrite(storage, "neg", (const void*)_neg->hist, CvAttrList());
      cvWrite(storage, "pos", (const void*)_pos->hist, CvAttrList());
      cvWrite(storage, "dif", (const void*)_dif->hist, CvAttrList());
      cvReleaseFileStorage(&storage);
      IOUT("Saved histograms to location \"" << _filename << "\".");
      break;
    }
    case 'l':
    {
      HistogramDetector::load();
      break;
    }
    default:
      return;
  }
}

bool HistogramDetector::detect (IplImage** imgs, std::vector<TapirObject>& objects)
{
  IplImage* const raw = imgs[0];
  if (raw == 0) return true;

  // update histograms

#if 0
  if (_first) {
    _dif->process(raw);
    _first = false;
  }
#endif

  if (_lock || _updates > 0) {
    _neg->process(raw);
    _neg->accumulate = 1;
    _updates--;
  }

  if (_first) {
    _dif->process(raw);
    if (_autoload) HistogramDetector::load();
    _first = false;
  }

  // find blobs

  _pos->clear();
  if (raw) _pos->process(raw);
  _pos->accumulate = 1;
  _dif->subtract_histogram(_pos->hist, _neg->hist, _dif->hist);
  cvCalcBackProject(_pos->planes, _detect, _dif->hist);

  cvLabel(_detect, _label, *_blobs);
  cvFilterByArea(*_blobs, _area[0], _area[1]);

  cvb::CvBlobs::iterator i;
  cvb::CvBlob *b;
  for (i=_blobs->begin(); i!=_blobs->end(); i++) {
    b = (*i).second;
    objects.push_back(TapirObject(b->centroid.x, b->centroid.y, b->area));
  }

  // generate output image

  cvCopyImage(raw, _canvas);
  cvRenderBlobs(_label, *_blobs, raw, _canvas, CV_BLOB_RENDER_CENTROID | CV_BLOB_RENDER_BOUNDING_BOX);

  imgs[_imgoffset] = _detect;
  imgs[_imgoffset+1] = _canvas;
  return true;
}

REGISTER(Detector, HistogramDetector, "Uses histogram differences to detect new objects.");

