#include "MOTLDDetector.h"
#include <cstring>

ObjectBox _box  = {0,0,0,0,0};

MOTLDDetector::MOTLDDetector ()
{
  _save = false;
  _load = false;
  _reset = false;
  _corner = false;
}

bool MOTLDDetector::init (const int width, const int height, const int imgs, unsigned int& add, const char *fname, const char* chapter)
{
  _width  = width;
  _height = height;
  _size   = _width * _height;
  _canvas = cvCreateImage(cvSize(_width, _height), IPL_DEPTH_8U, 3);
  _data   = new unsigned char[_size*3];

  add = 1;
  _imgoffset = imgs;

  ValueParser vp(fname, CHAPTER "Detector");
  vp.get("learning", _learning, true);
  vp.get("draw_mode", _drawmode, 255);

  _classfile = new char[255];
  if (vp.get("classifier", _classfile, 255) < 1)
    SET_STRING(_classfile, "classifier.moctld");
  
  bool load;
  vp.get("autoload", load, false);
  if (load) {
    _tld = new MultiObjectTLD(MultiObjectTLD::loadClassifier(_classfile));
  } else {
    bool color;
    vp.get("color", color, true);
    MOTLDSettings settings(color ? COLOR_MODE_RGB : COLOR_MODE_GRAY);
    settings.useColor = color;
    _tld = new MultiObjectTLD(_width, _height, settings);
  }

  REGISTER_KEY(MOTLDDetector, 'd', "Draw detections");
  REGISTER_KEY(MOTLDDetector, 't', "Draw crosses");
  REGISTER_KEY(MOTLDDetector, 'p', "Draw patches");
  REGISTER_KEY(MOTLDDetector, 'l', "Train classifier");
  REGISTER_KEY(MOTLDDetector, 'r', "Reset classifier");
  REGISTER_KEY(MOTLDDetector, 's', "Save classifier");
  REGISTER_KEY(MOTLDDetector, 'o', "Load classifier");
  REGISTER_HIDDEN_KEY(MOTLDDetector, 'L');
  REGISTER_HIDDEN_KEY(MOTLDDetector, 'R');

  return true;
}

void MOTLDDetector::deinit ()
{
  cvReleaseImage(&_canvas);
}

void MOTLDDetector::notify (const long int cmd, const char* params)
{
  switch (cmd) {
    case 'd': _drawmode ^= DEBUG_DRAW_DETECTIONS;  break;
    case 't': _drawmode ^= DEBUG_DRAW_CROSSES;  break;
    case 'p': _drawmode ^= DEBUG_DRAW_PATCHES;  break;
    case 'l':
      _learning = !_learning;
      IOUT("Learning " << (_learning ? "en" : "dis") << "abled");
      break;
    case 'r': _reset = true; break;
    case 's': _save = true;  break;
    case 'o': _load = true;  break;
    case 'L':
      strcpy(_buf, params);
#define PARSE_TOKEN(_dim_,_side_,_start_) \
      _token = strtok(_start_, " "); \
      if (_token == NULL) { \
        EOUT("Invalid parameters for event " << cmd << "."); \
        _corner = false; \
        return; \
      } \
      _coord = atoi(_token); \
      _box._side_ = abs(_box._dim_ - _coord); \
      if (!_corner || _coord < _box._dim_) \
        _box._dim_ = _coord;
      PARSE_TOKEN(x, width,  _buf);
      PARSE_TOKEN(y, height, NULL);
      if (_corner) {
        if(_box.width < 4 || _box.height < 4)
          EOUT("Bounding box too small!");
        else
          _tld->addObject(_box);
      }
      _corner = !_corner;
      break;
    case 'R':
      _corner = false;
      break;
    default:
      EOUT("Unhandled key-code: " << cmd);
      break;
  }
}

bool MOTLDDetector::detect (IplImage** imgs, std::vector<TapirObject>& objects)
{
  if (_reset) {
    _tld = new MultiObjectTLD(_width, _height, COLOR_MODE_RGB);
    _reset = false;
  }

  if (_load) {
    ifstream file(_classfile);
    if (file.good())
      _tld = new MultiObjectTLD(MultiObjectTLD::loadClassifier(_classfile));
    else
      EOUT("Cannot read file \"" << _classfile << "\".");
    _load = false;
  }

  for(_i=0; _i<_size; _i++) {
    _data[_i]         = imgs[0]->imageData[_i*3+2];
    _data[_i+_size]   = imgs[0]->imageData[_i*3+1];
    _data[_i+2*_size] = imgs[0]->imageData[_i*3];
  }
  _tld->processFrame(_data);
  _tld->getDebugImage(_data, _red, _green, _blue, _drawmode);
  for(_i=0; _i<_width*_height; ++_i){
    _canvas->imageData[3*_i+2] = _red.data()[_i];
    _canvas->imageData[3*_i+1] = _green.data()[_i];
    _canvas->imageData[3*_i+0] = _blue.data()[_i];
  }
  _tld->enableLearning(_learning);

  std::vector<ObjectBox> boxes = _tld->getObjectBoxes();
  std::vector<ObjectBox>::iterator it = boxes.begin();
  for (it=boxes.begin(); it<boxes.end(); it++) {
    ObjectBox box = *it;
    ///< \todo if not there, libMOTLD still reports confidence 50, even if loaded -> never seen before
    //int conf = _tld->getStatus(box.objectId);
    //conf = conf==STATUS_OK ? 100 : STATUS_LOST ? 0 : 50;
    //objects.push_back(TapirObject(box.x, box.y, box.width*box.height, 0, conf));
    if (_tld->getStatus(box.objectId) == STATUS_OK)
      objects.push_back(TapirObject(box.x+box.width/2, box.y+box.height/2, box.width*box.height));
  }

  if (_save) {
    _tld->saveClassifier((char*)_classfile);
    _save = false;
  }

  if (_corner)
    cvCircle(_canvas, cvPoint(_box.x,_box.y), 1, cvScalar(255,255,255), 2);

  imgs[_imgoffset] = _canvas;
  return true;
}

REGISTER(Detector, MOTLDDetector, "");

