#include "neuropredictor.h"
#include <cmath>

bool NeuroPredictor::init (const int width, const int height, const int imgs, unsigned int& add, const char* fname, const char* chapter)
{
  ValueParser vp(fname, CHAPTER "Tracker");
  vp.get("threshold", _threshold, 0);

  // get net location
  _filename = new char[255];
  if (vp.get("filename", _filename, 255) < 1) {
    EOUT("No filename specified.");
    return false;
  }

  // determine background
  char background[255];
  _use_map = _bin_map = false;
  if (vp.get("background", background, 255) < 1) {
    IOUT("No background specified. Using camera image.");
  } else if (strcmp(background, "binary") == 0) {
    _use_map = true;
    _bin_map = true;
  } else if (strcmp(background, "prediction") == 0) {
    _use_map = true;
  } else if (strcmp(background, "camera") != 0) {
    EOUT("Unknown background type \"" << background << "\".");
    return false;
  }
  vp.get("map_step", _step, 0.1);

  // determine patterns
  _patfile = new char[255];
  _use_patterns = vp.get("patterns", _patfile, 255) > 0;

  // determine cursor
  vp.get("binary_cursor", _bin_cursor, true);

  // data structires
  _canvas = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);
  _map    = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, 3);

  // counters
  _imgid = imgs;
  add = 1;
  _width  = width;
  _height = height;

  REGISTER_KEY(NeuroPredictor, 'r', "Reload network and redraw prediction map");
  return reload();
}


bool NeuroPredictor::reload ()
{
  // load network
  _net.load_net(_filename);
  if (_net.topo_data.out_count != 1 || _net.topo_data.in_count !=2) {
    EOUT("Cannot visualize network with topology " << _net.topo_data.in_count << " -> " << _net.topo_data.out_count << ", must be 2 -> 1");
    return false;
  }

  // draw prediction map
  if (_use_map) {
    cvRectangle(_map, cvPoint(0,0), cvPoint(_width,_height), cvScalar(0,0,0), CV_FILLED);
    double prediction, cx, cy, x, y;
    double xd = _width  * _step / 4.;
    double yd = _height * _step / 4.;
    for (x=-1.; x<=1.; x+=_step)
      for (y=-1; y<=1.; y+=_step) {
        cx = x * _width  / 2. + _width  / 2.;
        cy = y * _height / 2. + _height / 2.; 
        _net.in_vec[0] = x;
        _net.in_vec[1] = y;
        _net.forward_pass(_net.in_vec, _net.out_vec);
        if (_bin_map)
          _net.out_vec[0] = _net.out_vec[0] < _threshold ? -1. : 1.;
        prediction = 128 * _net.out_vec[0] + 128;
        cvRectangle(_map, cvPoint(cx-xd,cy-yd), cvPoint(cx+xd,cy+yd), cvScalar(prediction,prediction,prediction), CV_FILLED);
    }

    // draw patterns
    if (_use_patterns) {
      _patterns.load_pattern(_patfile); ///< \todo often causes crashes!
      for (_i=0; _i<_patterns.pattern_count; _i++) {
        prediction = _patterns.target[_i][0]==-1 ? 0 : 255;
        cx = _patterns.input[_i][0] * _width  / 2. + _width  / 2.;
        cy = _patterns.input[_i][1] * _height / 2. + _height / 2.;
        cvCircle(_map, cvPoint(cx,cy), 1, cvScalar(0,255-prediction,prediction), 2);
      }
    }
  }

  return true;
}

void NeuroPredictor::deinit ()
{
  cvReleaseImage(&_map);
  cvReleaseImage(&_canvas);
}

void NeuroPredictor::select (IplImage** imgs, std::vector<TapirObject>& objects)
{
  if (imgs[0] == 0) return;
  cvCopy(_use_map?_map:imgs[0], _canvas);
  imgs[_imgid] = _canvas;

  for (_i=0; _i<(int)objects.size(); _i++) {
    _net.in_vec[0] = 2. * objects.at(_i).x / _width  - 1.;
    _net.in_vec[1] = 2. * objects.at(_i).y / _height - 1.;
    _net.forward_pass(_net.in_vec, _net.out_vec);
    if (_bin_cursor)
      _net.out_vec[0] = _net.out_vec[0] < _threshold ? -1. : 1.;
    int prediction = 127 * _net.out_vec[0] * fabs(_net.out_vec[0]) + 128;
    cvCircle(_canvas, cvPoint(objects.at(_i).x,objects.at(_i).y), 5, cvScalar(0,255-prediction,prediction), 5);
  }
}

void NeuroPredictor::notify (const long int cmd, const char* params)
{
  if (cmd == 'r') reload();
}


REGISTER(Tracker, NeuroPredictor, "Gets a success prediction from a neural network.");
